from typing import TypedDict, Optional, Literal, Union, Dict, Any
import torch
from torch import nn
from ..snn._tt_infrastructure import TTModule, TTLayer
from .. import functional


class LIT(TTLayer, TTModule):
    def __init__(
            self,
            num_neurons: int,
            beta: Union[float, torch.Tensor] = 0.9,
            pos_threshold: Union[float, torch.Tensor] = 1.0,
            neg_threshold: Union[float, torch.Tensor] = 1.0,
            dim: int = -1,
            beta_rank: Literal[0, 1] = 1,
            pos_threshold_rank: Literal[0, 1] = 1,
            neg_threshold_rank: Literal[0, 1] = 1,
            learn_beta: bool = True,
            learn_pos_threshold: bool = True,
            learn_neg_threshold: bool = True,
            surrogate_derivative=functional.atan_surrogate(2.0),
    ):
        TTLayer.__init__(self, num_neurons, dim)
        TTModule.__init__(self)

        self._initialize_state("mem")
        self._register_decay("beta", beta, beta_rank, learn_beta)

        self.heaviside = surrogate_derivative
        self._register_threshold("pos_threshold", pos_threshold, pos_threshold_rank, learn_pos_threshold)
        self._register_threshold("neg_threshold", neg_threshold, neg_threshold_rank, learn_neg_threshold)

    def forward(self, x):
        self._ensure_states(x)

        x_moved = self._to_working_dim(x)

        mem_moved = self._to_working_dim(self.mem)
        mem_moved = mem_moved * self.beta + x_moved

        pos_spikes = self.heaviside(mem_moved - self.pos_threshold)
        neg_spikes = -self.heaviside(self.neg_threshold - mem_moved)
        spikes_moved = pos_spikes + neg_spikes
        spikes = self._from_working_dim(spikes_moved)

        self.mem = self._from_working_dim(mem_moved)

        return spikes


class DLIT(TTLayer, TTModule):
    def __init__(
            self,
            num_neurons: int,
            pos_beta: Union[float, torch.Tensor] = 0.9,
            neg_beta: Union[float, torch.Tensor] = 0.9,
            pos_threshold: Union[float, torch.Tensor] = 1.0,
            neg_threshold: Union[float, torch.Tensor] = 1.0,
            dim: int = -1,
            pos_beta_rank: Literal[0, 1] = 1,
            neg_beta_rank: Literal[0, 1] = 1,
            pos_threshold_rank: Literal[0, 1] = 1,
            neg_threshold_rank: Literal[0, 1] = 1,
            learn_pos_beta: bool = True,
            learn_neg_beta: bool = True,
            learn_pos_threshold: bool = True,
            learn_neg_threshold: bool = True,
            surrogate_derivative=functional.atan_surrogate(2.0),
    ):
        TTLayer.__init__(self, num_neurons, dim)
        TTModule.__init__(self)

        self._initialize_state("pos_mem")
        self._initialize_state("neg_mem")
        self._register_decay("pos_beta", pos_beta, pos_beta_rank, learn_pos_beta)
        self._register_decay("neg_beta", neg_beta, neg_beta_rank, learn_neg_beta)

        self.heaviside = surrogate_derivative
        self._register_threshold("pos_threshold", pos_threshold, pos_threshold_rank, learn_pos_threshold)
        self._register_threshold("neg_threshold", neg_threshold, neg_threshold_rank, learn_neg_threshold)

    def forward(self, x):
        self._ensure_states(x)

        x_moved = self._to_working_dim(x)

        pos_mem_moved = self._to_working_dim(self.pos_mem)
        neg_mem_moved = self._to_working_dim(self.neg_mem)

        pos_mem_moved = pos_mem_moved * self.pos_beta + torch.where(x_moved >= 0, x_moved, 0.0)
        neg_mem_moved = neg_mem_moved * self.neg_beta + torch.where(x_moved < 0, x_moved, 0.0)

        mem_moved = pos_mem_moved + neg_mem_moved

        pos_spikes = self.heaviside(mem_moved - self.pos_threshold)
        neg_spikes = -self.heaviside(self.neg_threshold - mem_moved)
        spikes_moved = pos_spikes + neg_spikes
        spikes = self._from_working_dim(spikes_moved)

        self.pos_mem = self._from_working_dim(pos_mem_moved)
        self.neg_mem = self._from_working_dim(neg_mem_moved)

        return spikes


class SLIT(TTLayer, TTModule):
    def __init__(
            self,
            num_neurons: int,
            alpha: Union[float, torch.Tensor] = 0.5,
            beta: Union[float, torch.Tensor] = 0.9,
            pos_threshold: Union[float, torch.Tensor] = 1.0,
            neg_threshold: Union[float, torch.Tensor] = 1.0,
            dim: int = -1,
            alpha_rank: Literal[0, 1] = 1,
            beta_rank: Literal[0, 1] = 1,
            pos_threshold_rank: Literal[0, 1] = 1,
            neg_threshold_rank: Literal[0, 1] = 1,
            learn_alpha: bool = True,
            learn_beta: bool = True,
            learn_pos_threshold: bool = True,
            learn_neg_threshold: bool = True,
            surrogate_derivative=functional.atan_surrogate(2.0),
    ):
        TTLayer.__init__(self, num_neurons, dim)
        TTModule.__init__(self)
        self._initialize_state("syn")
        self._register_decay("alpha", alpha, alpha_rank, learn_alpha)
        self._initialize_state("mem")
        self._register_decay("beta", beta, beta_rank, learn_beta)
        self.heaviside = surrogate_derivative
        self._register_threshold("pos_threshold", pos_threshold, pos_threshold_rank, learn_pos_threshold)
        self._register_threshold("neg_threshold", neg_threshold, neg_threshold_rank, learn_neg_threshold)

    def forward(self, x):
        self._ensure_states(x)

        x_moved = self._to_working_dim(x)

        syn_moved = self._to_working_dim(self.syn)
        syn_moved = syn_moved * self.alpha + x_moved * (1 - self.alpha)

        mem_moved = self._to_working_dim(self.mem)
        mem_moved = mem_moved * self.beta + syn_moved * (1 - self.beta)

        pos_spikes = self.heaviside(mem_moved - self.pos_threshold)
        neg_spikes = -self.heaviside(self.neg_threshold - mem_moved)
        spikes_moved = pos_spikes + neg_spikes
        spikes = self._from_working_dim(spikes_moved)

        self.syn = self._from_working_dim(spikes_moved)
        self.mem = self._from_working_dim(mem_moved)

        return spikes


class RLIT(TTLayer, TTModule):
    def __init__(
            self,
            num_neurons: int,
            beta: Union[float, torch.Tensor] = 0.9,
            gamma: Union[float, torch.Tensor] = 0.9,
            pos_threshold: Union[float, torch.Tensor] = 1.0,
            neg_threshold: Union[float, torch.Tensor] = 1.0,
            rec_weight: Union[float, torch.Tensor] = 0.0,
            bias: Union[float, torch.Tensor] = 0.0,
            dim: int = -1,
            beta_rank: Literal[0, 1] = 1,
            gamma_rank: Literal[0, 1] = 1,
            pos_threshold_rank: Literal[0, 1] = 1,
            neg_threshold_rank: Literal[0, 1] = 1,
            rec_weight_rank: Literal[0, 1] = 1,
            bias_rank: Literal[0, 1] = 1,
            learn_beta: bool = True,
            learn_gamma: bool = True,
            learn_pos_threshold: bool = True,
            learn_neg_threshold: bool = True,
            learn_rec_weight: bool = True,
            learn_bias: bool = True,
            surrogate_derivative=functional.atan_surrogate(2.0),
    ):
        TTLayer.__init__(self, num_neurons, dim)
        TTModule.__init__(self)

        self._initialize_state("mem")
        self._register_decay("beta", beta, beta_rank, learn_beta)

        self._initialize_state("rec")
        self._initialize_state("prev_output")
        self._register_decay("gamma", gamma, gamma_rank, learn_gamma)

        self.heaviside = surrogate_derivative
        self._register_threshold("pos_threshold", pos_threshold, pos_threshold_rank, learn_pos_threshold)
        self._register_threshold("neg_threshold", neg_threshold, neg_threshold_rank, learn_neg_threshold)

        self._register_parameter("rec_weight", rec_weight, rec_weight_rank, learn_rec_weight)
        self._register_parameter("bias", bias, bias_rank, learn_bias)

    def forward(self, x):
        self._ensure_states(x)

        x_moved = self._to_working_dim(x)

        rec_moved = self._to_working_dim(self.rec)
        prev_output_moved = self._to_working_dim(self.prev_output)
        rec_moved = rec_moved * self.gamma + prev_output_moved * (1 - self.gamma)

        mem_delta = rec_moved * self.rec_weight + x_moved + self.bias

        mem_moved = self._to_working_dim(self.mem)
        mem_moved = mem_moved * self.beta + mem_delta

        pos_spikes = self.heaviside(mem_moved - self.pos_threshold)
        neg_spikes = -self.heaviside(self.neg_threshold - mem_moved)
        spikes_moved = pos_spikes + neg_spikes
        spikes = self._from_working_dim(spikes_moved)

        self.mem = self._from_working_dim(mem_moved)
        self.rec = self._from_working_dim(rec_moved)
        self.prev_output = spikes

        return spikes



from tracetorch.snn.flex._leaky_integrator import LeakyIntegrator
from tracetorch.snn.flex._leaky_integrator import DEFAULT_ALPHA
from tracetorch.snn.flex._leaky_integrator import DEFAULT_BETA
# from ._leaky_integrator import DEFAULT_GAMMA
from tracetorch.snn.flex._leaky_integrator import DEFAULT_POS_THRESH
# from ._leaky_integrator import DEFAULT_NEG_THRESH
# from ._leaky_integrator import DEFAULT_POS_SCALE
# from ._leaky_integrator import DEFAULT_NEG_SCALE
# from ._leaky_integrator import DEFAULT_WEIGHT
# from ._leaky_integrator import DEFAULT_BIAS

from typing import Union, Literal, Any
import torch


class SLIF(LeakyIntegrator):
    def __init__(
            self,
            num_neurons: int,
            alpha: Union[float, torch.Tensor] = DEFAULT_ALPHA["value"],
            beta: Union[float, torch.Tensor] = DEFAULT_BETA["value"],
            pos_threshold: Union[float, torch.Tensor] = DEFAULT_POS_THRESH["value"],
            dim: int = -1,
            alpha_rank: Literal[0, 1] = DEFAULT_ALPHA["rank"],
            beta_rank: Literal[0, 1] = DEFAULT_BETA["rank"],
            pos_threshold_rank: Literal[0, 1] = DEFAULT_POS_THRESH["rank"],
            ema_alpha: bool = DEFAULT_ALPHA["ema"],
            ema_beta: bool = DEFAULT_BETA["ema"],
            learn_alpha: bool = DEFAULT_ALPHA["learnable"],
            learn_beta: bool = DEFAULT_BETA["learnable"],
            learn_pos_threshold: bool = DEFAULT_POS_THRESH["learnable"],
            surrogate_derivative: Any = DEFAULT_POS_THRESH["surrogate"],
    ):
        alpha_setup = {
            "value": alpha,
            "rank": alpha_rank,
            "ema": ema_alpha,
            "learnable": learn_alpha,
        }

        beta_setup = {
            "value": beta,
            "rank": beta_rank,
            "ema": ema_beta,
            "learnable": learn_beta,
        }

        pos_threshold_setup = {
            "value": pos_threshold,
            "rank": pos_threshold_rank,
            "surrogate": surrogate_derivative,
            "learnable": learn_pos_threshold,
        }

        super().__init__(
            num_neurons=num_neurons,
            dim=dim,
            alpha_pos_setup=alpha_setup,
            beta_pos_setup=beta_setup,
            pos_threshold_setup=pos_threshold_setup,
        )



from ._tt_infrastructure import TTLayer, TTModule

# from ._li_layers import LI, DLI, SLI, RLI, DSLI, DRLI, SRLI, DSRLI
# from ._lib_layers import LIB, DLIB, SLIB, RLIB, DSLIB, DRLIB, SRLIB, DSRLIB
# from ._lit_layers import LIT, DLIT, SLIT, RLIT, DSLIT, DRLIT, SRLIT, DSRLIT
# from ._lits_layers import LITS, DLITS, SLITS, RLITS, DSLITS, DRLITS, SRLITS, DSRLITS

# import flex

from ._li_layers import LI, DLI, SLI, RLI
from ._lib_layers import LIB, DLIB, SLIB, RLIB
from ._lit_layers import LIT, DLIT, SLIT, RLIT
from ._lits_layers import LITS, DLITS, SLITS, RLITS



import torch
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import cm
from matplotlib.colors import Normalize


def render_image(tensor, figsize=(4, 4), title=None, vmin=None, vmax=None):
    """
    Plot a tensor image.

    - tensor: torch.Tensor in either HW (H,W) or CHW (C,H,W) format.
              C may be 1 (grayscale) or 3 (RGB).
    - figsize: matplotlib figure size tuple
    - title: optional title string
    - vmin/vmax: if provided, these define the original data range (used for the colorbar).
                 If None, the function uses tensor.min() / tensor.max().
    Behavior change for grayscale: the image is rescaled to [0,1] for display (max contrast),
    but the colorbar shows the original numeric range (vmin..vmax).
    :param tensor:
    :param figsize:
    :param title:
    :param vmin:
    :param vmax:
    :return:
    """
    if not torch.is_tensor(tensor):
        tensor = torch.as_tensor(tensor)

    t = tensor.detach()
    if t.device.type != "cpu":
        t = t.cpu()

    # helper to determine original vmin/vmax (either from args or data)
    def _orig_range(arr, vmin_arg, vmax_arg):
        data_min = float(np.nanmin(arr))
        data_max = float(np.nanmax(arr))
        orig_vmin = data_min if vmin_arg is None else float(vmin_arg)
        orig_vmax = data_max if vmax_arg is None else float(vmax_arg)
        return orig_vmin, orig_vmax

    if t.dim() == 2:
        # HW grayscale
        img = t.numpy().astype(np.float32)

        orig_vmin, orig_vmax = _orig_range(img, vmin, vmax)

        # avoid divide-by-zero for constant images
        if orig_vmax > orig_vmin:
            img_scaled = (img - orig_vmin) / (orig_vmax - orig_vmin)
            # numerical safety
            img_scaled = np.clip(img_scaled, 0.0, 1.0)
        else:
            # constant image -> show mid-gray but colorbar will show the constant value
            img_scaled = np.full_like(img, 0.5, dtype=np.float32)

        fig, ax = plt.subplots(figsize=figsize)
        im = ax.imshow(img_scaled, cmap="gray", vmin=0.0, vmax=1.0, interpolation="nearest")
        ax.axis("off")
        if title:
            ax.set_title(title)

        # colorbar referencing original scale
        mappable = cm.ScalarMappable(norm=Normalize(vmin=orig_vmin, vmax=orig_vmax), cmap="gray")
        mappable.set_array([])  # required for colorbar
        plt.colorbar(mappable, ax=ax, fraction=0.046, pad=0.04)
        plt.show()
        return

    if t.dim() == 3:
        c, h, w = t.shape
        if c == 1:
            img = t.squeeze(0).numpy().astype(np.float32)

            orig_vmin, orig_vmax = _orig_range(img, vmin, vmax)

            if orig_vmax > orig_vmin:
                img_scaled = (img - orig_vmin) / (orig_vmax - orig_vmin)
                img_scaled = np.clip(img_scaled, 0.0, 1.0)
            else:
                img_scaled = np.full_like(img, 0.5, dtype=np.float32)

            fig, ax = plt.subplots(figsize=figsize)
            im = ax.imshow(img_scaled, cmap="gray", vmin=0.0, vmax=1.0, interpolation="nearest")
            ax.axis("off")
            if title:
                ax.set_title(title)

            mappable = cm.ScalarMappable(norm=Normalize(vmin=orig_vmin, vmax=orig_vmax), cmap="gray")
            mappable.set_array([])
            plt.colorbar(mappable, ax=ax, fraction=0.046, pad=0.04)
            plt.show()
            return

        elif c == 3:
            # CHW -> HWC for matplotlib (leave RGB untouched)
            img = t.permute(1, 2, 0).numpy()
            img = np.clip(img, 0.0, 1.0)  # assume RGB in 0..1; you can change handling if needed
            fig, ax = plt.subplots(figsize=figsize)
            ax.imshow(img, interpolation="nearest")
            ax.axis("off")
            if title:
                ax.set_title(title)
            plt.show()
            return
        else:
            raise ValueError(f"Unsupported channel dimension: C={c}. Expected C==1 or C==3 for CHW.")

    raise ValueError(f"Unsupported tensor shape {tuple(t.shape)}. Expected HW or CHW.")



from tracetorch.snn.flex._leaky_integrator import LeakyIntegrator
# from ._leaky_integrator import DEFAULT_ALPHA
from tracetorch.snn.flex._leaky_integrator import DEFAULT_BETA
from tracetorch.snn.flex._leaky_integrator import DEFAULT_GAMMA
# from ._leaky_integrator import DEFAULT_POS_THRESH
# from ._leaky_integrator import DEFAULT_NEG_THRESH
# from ._leaky_integrator import DEFAULT_POS_SCALE
# from ._leaky_integrator import DEFAULT_NEG_SCALE
from tracetorch.snn.flex._leaky_integrator import DEFAULT_REC_WEIGHT
from tracetorch.snn.flex._leaky_integrator import DEFAULT_BIAS

from typing import Union, Literal
import torch


class RReadout(LeakyIntegrator):
    def __init__(
            self,
            num_neurons: int,
            beta: Union[float, torch.Tensor] = DEFAULT_BETA["value"],
            gamma: Union[float, torch.Tensor] = DEFAULT_GAMMA["value"],
            rec_weight: Union[float, torch.Tensor] = DEFAULT_REC_WEIGHT["value"],
            bias: Union[float, torch.Tensor] = DEFAULT_BIAS["value"],
            dim: int = -1,
            beta_rank: Literal[0, 1] = DEFAULT_BETA["rank"],
            gamma_rank: Literal[0, 1] = DEFAULT_GAMMA["rank"],
            rec_weight_rank: Literal[0, 1] = DEFAULT_REC_WEIGHT["rank"],
            bias_rank: Literal[0, 1] = DEFAULT_BIAS["rank"],
            ema_gamma: bool = DEFAULT_GAMMA["ema"],
            learn_beta: bool = DEFAULT_BETA["learnable"],
            learn_gamma: bool = DEFAULT_GAMMA["learnable"],
            learn_rec_weight: bool = DEFAULT_REC_WEIGHT["learnable"],
            learn_bias: bool = DEFAULT_BIAS["learnable"],
    ):
        beta_setup = {
            "value": beta,
            "rank": beta_rank,
            "ema": True,
            "learnable": learn_beta,
        }

        gamma_setup = {
            "value": gamma,
            "rank": gamma_rank,
            "ema": ema_gamma,
            "learnable": learn_gamma,
        }

        rec_weight_setup = {
            "value": rec_weight,
            "rank": rec_weight_rank,
            "learnable": learn_rec_weight,
        }

        bias_setup = {
            "value": bias,
            "rank": bias_rank,
            "learnable": learn_bias,
        }

        super().__init__(
            num_neurons=num_neurons,
            dim=dim,
            beta_pos_setup=beta_setup,
            gamma_pos_setup=gamma_setup,
            rec_weight_setup=rec_weight_setup,
            bias_setup=bias_setup,
        )



from tracetorch.snn.flex._leaky_integrator import LeakyIntegrator
# from ._leaky_integrator import DEFAULT_ALPHA
from tracetorch.snn.flex._leaky_integrator import DEFAULT_BETA
# from ._leaky_integrator import DEFAULT_GAMMA
# from ._leaky_integrator import DEFAULT_POS_THRESH
# from ._leaky_integrator import DEFAULT_NEG_THRESH
# from ._leaky_integrator import DEFAULT_POS_SCALE
# from ._leaky_integrator import DEFAULT_NEG_SCALE
# from ._leaky_integrator import DEFAULT_WEIGHT
# from ._leaky_integrator import DEFAULT_BIAS

from typing import Union, Literal
import torch


class Readout(LeakyIntegrator):
	def __init__(
			self,
			num_neurons: int,
			beta: Union[float, torch.Tensor] = DEFAULT_BETA["value"],
			dim: int = -1,
			beta_rank: Literal[0, 1] = DEFAULT_BETA["rank"],
			learn_beta: bool = DEFAULT_BETA["learnable"],
	):
		beta_setup = {
			"value": beta,
			"rank": beta_rank,
			"ema": True,
			"learnable": learn_beta,
		}

		super().__init__(
			num_neurons=num_neurons,
			dim=dim,
			beta_pos_setup=beta_setup,
		)



from tracetorch.snn.flex._leaky_integrator import LeakyIntegrator
from tracetorch.snn.flex._leaky_integrator import DEFAULT_ALPHA
from tracetorch.snn.flex._leaky_integrator import DEFAULT_BETA
from tracetorch.snn.flex._leaky_integrator import DEFAULT_GAMMA
from tracetorch.snn.flex._leaky_integrator import DEFAULT_POS_THRESH
# from ._leaky_integrator import DEFAULT_NEG_THRESH
# from ._leaky_integrator import DEFAULT_POS_SCALE
# from ._leaky_integrator import DEFAULT_NEG_SCALE
from tracetorch.snn.flex._leaky_integrator import DEFAULT_REC_WEIGHT
from tracetorch.snn.flex._leaky_integrator import DEFAULT_BIAS

from typing import Union, Literal, Any
import torch


class SRLIF(LeakyIntegrator):
	def __init__(
			self,
			num_neurons: int,
			alpha: Union[float, torch.Tensor] = DEFAULT_ALPHA["value"],
			beta: Union[float, torch.Tensor] = DEFAULT_BETA["value"],
			gamma: Union[float, torch.Tensor] = DEFAULT_GAMMA["value"],
			pos_threshold: Union[float, torch.Tensor] = DEFAULT_POS_THRESH["value"],
			rec_weight: Union[float, torch.Tensor] = DEFAULT_REC_WEIGHT["value"],
			bias: Union[float, torch.Tensor] = DEFAULT_BIAS["value"],
			dim: int = -1,
			alpha_rank: Literal[0, 1] = DEFAULT_ALPHA["rank"],
			beta_rank: Literal[0, 1] = DEFAULT_BETA["rank"],
			gamma_rank: Literal[0, 1] = DEFAULT_GAMMA["rank"],
			pos_threshold_rank: Literal[0, 1] = DEFAULT_POS_THRESH["rank"],
			rec_weight_rank: Literal[0, 1] = DEFAULT_REC_WEIGHT["rank"],
			bias_rank: Literal[0, 1] = DEFAULT_BIAS["rank"],
			ema_alpha: bool = DEFAULT_ALPHA["ema"],
			ema_beta: bool = DEFAULT_BETA["ema"],
			ema_gamma: bool = DEFAULT_GAMMA["ema"],
			learn_alpha: bool = DEFAULT_ALPHA["learnable"],
			learn_beta: bool = DEFAULT_BETA["learnable"],
			learn_gamma: bool = DEFAULT_GAMMA["learnable"],
			pos_learn_threshold: bool = DEFAULT_POS_THRESH["learnable"],
			learn_rec_weight: bool = DEFAULT_REC_WEIGHT["learnable"],
			learn_bias: bool = DEFAULT_BIAS["learnable"],
			surrogate_derivative: Any = DEFAULT_POS_THRESH["surrogate"],
	):
		alpha_setup = {
			"value": alpha,
			"rank": alpha_rank,
			"ema": ema_alpha,
			"learnable": learn_alpha,
		}

		beta_setup = {
			"value": beta,
			"rank": beta_rank,
			"ema": ema_beta,
			"learnable": learn_beta,
		}

		gamma_setup = {
			"value": gamma,
			"rank": gamma_rank,
			"ema": ema_gamma,
			"learnable": learn_gamma,
		}

		pos_threshold_setup = {
			"value": pos_threshold,
			"rank": pos_threshold_rank,
			"surrogate": surrogate_derivative,
			"learnable": pos_learn_threshold,
		}

		rec_weight_setup = {
			"value": rec_weight,
			"rank": rec_weight_rank,
			"learnable": learn_rec_weight,
		}

		bias_setup = {
			"value": bias,
			"rank": bias_rank,
			"learnable": learn_bias,
		}

		super().__init__(
			num_neurons=num_neurons,
			dim=dim,
			alpha_pos_setup=alpha_setup,
			beta_pos_setup=beta_setup,
			gamma_pos_setup=gamma_setup,
			pos_threshold_setup=pos_threshold_setup,
			rec_weight_setup=rec_weight_setup,
			bias_setup=bias_setup,
		)



from typing import TypedDict, Optional, Literal, Union, Dict, Any, Set
import torch
from torch import nn
from .. import functional


class TTLayer:
    """A mixin helper class, used to help manage parameters and hidden states."""

    def __init__(self, num_neurons: int, dim: int = -1):
        self._state_names = set()
        self.num_neurons = num_neurons
        self.dim = dim

    def _register_parameter(
            self,
            name: str,
            value: Union[float, torch.Tensor],
            rank: Literal[0, 1],
            learnable: bool,
            inverse_function=lambda x: x,
            activation_function=lambda x: x,
    ):
        if isinstance(value, torch.Tensor):
            if value.ndim == 0:
                pass
            elif value.ndim == 1:
                assert value.numel() == self.num_neurons, f"{name} does not have {self.num_neurons} elements"
            else:
                raise ValueError(f"rank (.ndim) of provided {name} is not 0 (scalar) or 1 (vector)")
            param_tensor = inverse_function(value)
        else:
            value = float(value)
            if rank == 0:
                param_tensor = inverse_function(torch.tensor(value))
            elif rank == 1:
                param_tensor = inverse_function(torch.full([self.num_neurons], value))
            else:
                raise ValueError(f"{name} rank is not 0 (scalar) or 1 (vector)")

        # save the raw_ variant of the parameter as an nn.Parameter or buffer
        if learnable:
            setattr(self, f"raw_{name}", nn.Parameter(param_tensor.detach().clone()))
        else:
            self.register_buffer(f"raw_{name}", param_tensor.detach().clone())

        # create a @property of the raw_ parameter that passes it through the respective activation function
        setattr(self.__class__, name, self._make_property(name, activation_function))

    @staticmethod
    def _make_property(name, activation_function):
        def getter(self):
            return activation_function(getattr(self, f"raw_{name}"))

        return property(getter)

    def _register_decay(
            self,
            name: str,
            value: Union[float, torch.Tensor],
            rank: Literal[0, 1],
            learnable: bool,
    ):
        self._register_parameter(
            name,
            value,
            rank,
            learnable,
            inverse_function=functional.sigmoid_inverse,
            activation_function=nn.functional.sigmoid,
        )

    def _register_threshold(
            self,
            name: str,
            value: Union[float, torch.Tensor],
            rank: Literal[0, 1],
            learnable: bool,
    ):
        self._register_parameter(
            name,
            value,
            rank,
            learnable,
            inverse_function=functional.softplus_inverse,
            activation_function=nn.functional.softplus,
        )

    def _initialize_state(self, state_name: str):
        """Initialize and register a state name for bulk operations"""
        self._state_names.add(state_name)
        setattr(self, state_name, None)

    def _detach_state(self, state_name: str):
        """Detach a state tensor if it exists"""
        state = getattr(self, state_name)
        if state is not None:
            setattr(self, state_name, state.detach())

    def detach_states(self):
        """Detach all registered states"""
        for state_name in self._state_names:
            self._detach_state(state_name)

    def _zero_state(self, state_name: str):
        """Set a state to None"""
        setattr(self, state_name, None)

    def zero_states(self):
        """Zero all registered states"""
        for state_name in self._state_names:
            self._zero_state(state_name)

    def _ensure_state(self, state_name: str, reference_tensor: torch.Tensor):
        """Initialize a state with zeros_like if None, otherwise return existing state"""
        state = getattr(self, state_name)
        if state is None:
            state = torch.zeros_like(reference_tensor)
            setattr(self, state_name, state)

    def _ensure_states(self, reference_tensor: torch.Tensor):
        """Ensure all registered states are initialized"""
        for state_name in self._state_names:
            self._ensure_state(state_name, reference_tensor)

    def _to_working_dim(self, tensor: torch.Tensor) -> torch.Tensor:
        """Move tensor to working dimension (last dim)"""
        return tensor.movedim(self.dim, -1)

    def _from_working_dim(self, tensor: torch.Tensor) -> torch.Tensor:
        """Move tensor back from working dimension"""
        return tensor.movedim(-1, self.dim)


class TTModule(nn.Module):
    """
    Base class that makes it trivial to call lifecycle methods (zero_states, detach_states, ...)
    across an entire model tree (registered submodules *and* common non-module containers).
    Inherit your models from this to get model.zero_states() / model.detach_states() behavior.
    """

    def get_param_count(self) -> int:
        total_params = sum(p.numel() for p in self.parameters())
        return total_params

    def zero_states(self) -> None:
        """Public API — zero any stateful child that implements zero_states()."""
        self._call_recursive("zero_states")

    def detach_states(self) -> None:
        """Public API — detach (stop-gradient) any stateful child that implements detach_states()."""
        self._call_recursive("detach_states")

    # --- internal recursive walker ---
    def _call_recursive(self, method_name: str) -> None:
        """
        Walk the object graph rooted at self and call `method_name()` on any object
        that exposes it. Uses a visited set to avoid duplicates / infinite loops.
        """
        visited: Set[int] = set()

        def recurse(obj: Any) -> None:
            oid = id(obj)
            if oid in visited:
                return
            visited.add(oid)

            # 1) If object defines the requested method, call it.
            #    If object is an instance of TTModule, we call the override (if any).
            #    If object is a plain python object with a callable method_name, call that too.
            fn = getattr(obj, method_name, None)
            if callable(fn):
                # Avoid calling TTModule.zero_states itself in a naive way that would loop forever:
                # - If obj is a TTModule instance, check whether its class actually overrides the method
                #   (i.e., the implementation is not the same as TTModule.method_name).
                if isinstance(obj, TTModule):
                    # get the unbound function defined on the class (if any)
                    cls_fn = getattr(obj.__class__, method_name, None)
                    base_fn = getattr(TTModule, method_name, None)
                    if cls_fn is not None and cls_fn is not base_fn:
                        # the class overrides the method -> call the override
                        try:
                            fn()
                        except TypeError:
                            # defensive: some implementations may accept args; ignore them here
                            fn()
                # else: class doesn't override, so don't call TTModule.zero_states (that would just re-enter recursion)
                else:
                    # Non-TTModule object that happens to have a method_name: call it
                    try:
                        fn()
                    except TypeError:
                        fn()

            # 2) Recurse into registered submodules (this covers ModuleList, Sequential, etc.)
            if isinstance(obj, nn.Module):
                for child in obj._modules.values():
                    if child is None:
                        continue
                    recurse(child)

            # 3) Recurse into container attributes that might hold modules or other objects with methods.
            #    This covers plain python lists/tuples/dicts/sets assigned as attributes on modules.
            #    We purposely ignore common atomic types (tensors, numbers, strings).
            try:
                attrs = getattr(obj, "__dict__", {})
            except Exception:
                attrs = {}

            for attr in attrs.values():
                if attr is None:
                    continue
                # common containers
                if isinstance(attr, (list, tuple, set)):
                    for el in attr:
                        recurse(el)
                elif isinstance(attr, dict):
                    for el in attr.values():
                        recurse(el)
                else:
                    # If attribute is a module or object, recurse (visited prevents duplicates)
                    if isinstance(attr, (nn.Module, object)):
                        recurse(attr)

        # start recursion from self
        recurse(self)
